from django.conf import settings
from pusher import Pusher


def send_pusher_event(
    signal,
    sender,
    instance,
    channel,
    event_name,
    data,
    socket_id=None,
    **kwargs,
):
    """
    Send a pusher event from a signal
    """
    try:
        pusher_cluster = settings.PUSHER_CLUSTER
    except AttributeError:
        pusher_cluster = 'mt1'

    pusher = Pusher(
        app_id=settings.PUSHER_APP_ID,
        key=settings.PUSHER_KEY,
        secret=settings.PUSHER_SECRET,
        cluster=pusher_cluster,
    )
    pusher.trigger(
        [channel],
        event_name,
        data,
    )



"""
Mixin Classes for Pusher integration with Views
"""
from collections import defaultdict

from drf_model_pusher.signals import view_pre_destroy, view_post_save

pusher_backend_registry = defaultdict(list)


class PusherBackendMetaclass(type):
    """
    Register PusherBackend's with a registry for model lookups, supports
    abstract classes
    """

    def __new__(mcs, cls, bases, dicts):
        if dicts.get('Meta') and hasattr(dicts.get('Meta'), 'abstract'):
            dicts['__metaclass__'] = mcs
            return super().__new__(mcs, cls, bases, dicts)

        assert (
            dicts.get('serializer_class', None) is not None,
            'PusherBackends require a serializer_class'
        )
        dicts['__metaclass__'] = mcs

        final_cls = super().__new__(mcs, cls, bases, dicts)

        model_name = dicts['serializer_class'].Meta.model.__name__.lower()
        pusher_backend_registry[model_name].append(final_cls)
        return final_cls


class PusherBackend(metaclass=PusherBackendMetaclass):
    """
    PusherBackend is the base class for implementing serializers with Pusher
    """

    class Meta:
        abstract = True

    def __init__(self, view):
        self.view = view
        self.pusher_socket_id = self.get_pusher_socket(view)

    def get_pusher_socket(self, view):
        pusher_socket = view.request.META.get("HTTP_X_PUSHER_SOCKET_ID", None)
        return pusher_socket

    def push_change(self, event, instance=None, pre_destroy=False, ignore=False):
        channel, event_name, data = self.get_packet(event, instance)
        if pre_destroy:
            view_pre_destroy.send(
                sender=self.__class__,
                instance=self,
                channel=channel,
                event_name=event_name,
                data=data,
                socket_id=self.pusher_socket_id if ignore else None,
            )
        else:
            view_post_save.send(
                sender=self.__class__,
                instance=self,
                channel=channel,
                event_name=event_name,
                data=data,
                socket_id=self.pusher_socket_id if ignore else None,
            )

    def get_event_name(self, event_type):
        """Return the model name and the event_type separated by a dot"""
        serializer_class = self.get_serializer_class()
        return f"{serializer_class.Meta.model.__name__.lower()}.{event_type}"

    def get_serializer_class(self):
        """Return the views serializer class"""
        return self.view.get_serializer_class()

    def get_serializer(self, view, *args, **kwargs):
        """Return the serializer initialized with the views serializer context"""
        serializer_class = self.get_serializer_class()
        kwargs["context"] = view.get_serializer_context()
        return serializer_class(*args, **kwargs)

    def get_channel(self, instance=None):
        """Return the channel from the view or instance"""
        channel = self.view.get_pusher_channel()
        return channel

    def get_packet(self, event, instance):
        """Return a tuple consisting of the channel, event name, and the JSON serializable data."""
        channel = self.get_channel(instance=instance)
        event_name = self.get_event_name(event)
        data = self.get_serializer(self.view, instance=instance).data
        return channel, event_name, data


class PrivatePusherBackend(PusherBackend):
    """PrivatePusherBackend is the base class for implementing serializers
    with Pusher and prefixing the channel with `private-`."""

    class Meta:
        abstract = True

    def get_channel(self, instance=None):
        """Return the channel prefixed with `private-`"""
        channel = super().get_channel(instance=instance)
        return f"private-{channel}"


def get_models_pusher_backends(model):
    """Return the pusher backends registered for a model"""
    return pusher_backend_registry.get(model.__name__.lower(), [])



from drf_model_pusher.backends import get_models_pusher_backends
from drf_model_pusher.exceptions import ModelPusherException
from drf_model_pusher.signals import view_post_save


class ModelPusherViewMixin(object):
    """Enables views to push changes through pusher"""

    pusher_backends = []

    PUSH_CREATE = 'create'
    PUSH_UPDATE = 'update'
    PUSH_DELETE = 'delete'

    def __init__(self, push_creations=True, push_updates=True, push_deletions=True, **kwargs):
        self.push_creations = push_creations
        self.push_updates = push_updates
        self.push_deletions = push_deletions
        super().__init__(**kwargs)
        self.pusher_backends = self.get_models_pusher_backends()

    def get_models_pusher_backends(self):
        if not hasattr(self, "queryset"):
            raise ModelPusherException("View must have a queryset defined")
        return get_models_pusher_backends(self.queryset.model)

    def get_pusher_channel(self):
        """Return the channel from the view"""
        raise NotImplementedError(
            f"{self.__class__.__name__} must implement the `get_pusher_channel` method to use the PushChangesMixin"
        )

    def get_pusher_backends(self):
        """Return all the pusher backends registered for this views model"""
        return [pusher_backend(view=self) for pusher_backend in self.pusher_backends]

    def push_changes(self, event=PUSH_UPDATE, instance=None, pre_destroy=False):
        """Triggers the push_change method for all the pusher backends registered on this views model"""
        for pusher_backend in self.get_pusher_backends():
            pusher_backend.push_change(event, instance, pre_destroy=pre_destroy)

    def perform_update(self, serializer):
        """Update the object and then send the pusher event"""
        super().perform_update(serializer)
        if self.push_updates:
            self.push_changes(self.PUSH_UPDATE, serializer.instance)

    def perform_create(self, serializer):
        """Create the object and then send the pusher event"""
        super().perform_create(serializer)
        if self.push_creations:
            print(self.get_pusher_backends())
            self.push_changes(self.PUSH_CREATE, serializer.instance)

    def perform_destroy(self, instance):
        """Sends the pusher event and then destroys the object

        This is done to ensure the serialization can occur"""
        if self.push_deletions:
            self.push_changes(self.PUSH_DELETE, instance, pre_destroy=True)
        super().perform_destroy(instance)

    def push(self, channel, event_name, data):
        """Dispatch arbitrary data

        Intended to be invoked directly for custom purposes"""
        view_post_save.send(
            instance=self,
            sender=self.__class__,
            channel=channel,
            event_name=event_name,
            data=data
        )



from django.dispatch import Signal


view_post_save = Signal(
    providing_args=['instance', 'channel', 'event_name', 'data', 'socket_id']
)

view_pre_destroy = Signal(
    providing_args=['instance', 'channel', 'event_name', 'data', 'socket_id']
)



from drf_model_pusher.receivers import send_pusher_event
from drf_model_pusher.signals import view_post_save, view_pre_destroy


def connect_pusher_views():
    """
    Register the send_pusher_event with the view_post_save and
    view_pre_destroy signals
    """
    view_post_save.connect(send_pusher_event)
    view_pre_destroy.connect(send_pusher_event)
